import json

from django_unused2.dataclasses import AnalysisResult


def generate_dot(analysis_result: AnalysisResult) -> str:
    lines = ["digraph G {"]

    # Add nodes for Python files and Templates
    for py in analysis_result.python_files:
        lines.append(f'    "{py.id}" [label="{py.relative_path}", shape=box];')
    for tmpl in analysis_result.templates:
        lines.append(f'    "{tmpl.id}" [label="{tmpl.relative_path}", shape=ellipse];')

    # Add edges for references
    for ref in analysis_result.references:
        color = "red" if ref.broken else "black"
        lines.append(
            f'    "{ref.source_id}" -> "{ref.target_id}" [label="{ref.reference_type.name}", color="{color}"];'
        )

    lines.append("}")

    return "\n".join(lines)


def generate_cytoscape_json(analysis_result: AnalysisResult) -> str:
    elements = []
    never_used_ids = {t.id for t in analysis_result.never_referenced_templates}

    node_ids = set()

    for ref in analysis_result.references:
        if ref.broken:
            continue
        node_ids.add(ref.source_id)
        node_ids.add(ref.target_id)

    for ref in analysis_result.references:
        if ref.broken:
            continue
        if ref.source_id not in never_used_ids and ref.target_id not in never_used_ids:
            elements.append(
                {
                    "data": {
                        "source": ref.source_id,
                        "target": ref.target_id,
                        "label": ref.reference_type.name,
                        "line": ref.line,
                    }
                }
            )

    for ref in analysis_result.references:
        if ref.broken:
            continue
        if not (
            ref.source_id not in never_used_ids and ref.target_id not in never_used_ids
        ):
            elements.append(
                {
                    "data": {
                        "source": ref.source_id,
                        "target": ref.target_id,
                        "label": ref.reference_type.name,
                        "line": ref.line,
                    }
                }
            )

    for py in analysis_result.python_files:
        if py.id not in node_ids:
            continue
        app = ""
        if py.app_config:
            app = py.app_config.name

        elements.append(
            {
                "data": {
                    "id": py.id,
                    "label": py.relative_path,
                    "type": "python",
                    "app": app,
                }
            }
        )

    for tmpl in analysis_result.templates:
        if tmpl.id not in node_ids:
            continue
        used = tmpl.id not in never_used_ids
        app = ""
        if tmpl.app_config:
            app = tmpl.app_config.name
        elements.append(
            {
                "data": {
                    "id": tmpl.id,
                    "label": tmpl.relative_path,
                    "type": "template",
                    "app": app,
                    "used": used,
                }
            }
        )

    return json.dumps(elements, indent=4)



from typing import List, Optional, Dict, Set

from colorama import Fore

from django_unused2.dataclasses import (
    TemplateFilterOptions,
    Template,
    TemplateReference,
    Python,
    AnalysisResult,
)
from django_unused2.file_finder import (
    find_global_templates,
    find_app_templates,
    find_py_files,
    find_all_references,
)


def filter_templates(
    templates: List[Template], filter_options: Optional[TemplateFilterOptions]
) -> List[Template]:
    if filter_options and filter_options.excluded_apps:
        excluded_apps_set = set(filter_options.excluded_apps)
        templates = [
            t
            for t in templates
            if not (t.app_config and t.app_config.name in excluded_apps_set)
        ]

    if filter_options and filter_options.excluded_template_dirs:
        excluded_dirs_set = set(filter_options.excluded_template_dirs)
        templates = [
            t
            for t in templates
            if not any(t.relative_path.startswith(d) for d in excluded_dirs_set)
        ]

    if filter_options and filter_options.excluded_templates:
        et = set(filter_options.excluded_templates)
        templates = [
            template
            for template in templates
            if template.relative_path not in et and template.absolute_path not in et
        ]

    return templates


def analyze_references(
    references: List[TemplateReference],
    templates: List[Template],
    python_files: List[Python],
) -> AnalysisResult:
    broken_references = find_broken_references(references, python_files, templates)
    never_referenced_templates = find_unreferenced_templates(
        references, templates, python_files
    )

    return AnalysisResult(
        never_referenced_templates=never_referenced_templates,
        broken_references=broken_references,
        references=references,
        python_files=python_files,
        templates=templates,
    )


def find_unreferenced_templates(
    references: List[TemplateReference],
    templates: List[Template],
    python_files: List[Python],
) -> List[Template]:
    template_dict: Dict[str, Template] = {
        template.id: template for template in templates
    }
    python_file_dict: Dict[str, Python] = {
        py_file.id: py_file for py_file in python_files
    }

    visited_templates: Set[str] = set()

    for ref in references:
        if ref.source_id in python_file_dict and ref.target_id in template_dict:
            visited_templates.add(ref.target_id)

    for ref in references:
        if ref.source_id in visited_templates and ref.target_id in template_dict:
            visited_templates.add(ref.target_id)

    for ref in references:
        if ref.source_id in visited_templates and ref.target_id in template_dict:
            visited_templates.add(ref.target_id)

    for ref in references:
        if ref.source_id in visited_templates and ref.target_id in template_dict:
            visited_templates.add(ref.target_id)

    never_visited = [
        t for t in templates if t.id not in visited_templates and t.local_app
    ]

    return never_visited


def find_broken_references(
    references: List[TemplateReference],
    python_files: List[Python],
    templates: List[Template],
) -> List[TemplateReference]:
    local_ids = set([t.id for t in templates if t.local_app])
    return [r for r in references if r.broken and r.source_id in local_ids]


def run_analysis(config: TemplateFilterOptions) -> AnalysisResult:
    templates = filter_templates(find_app_templates() + find_global_templates(), config)
    python_files = find_py_files()
    references = find_all_references(templates, python_files)
    return analyze_references(references, templates, python_files)



import os
from typing import List, Dict

from colorama import init, Fore

from django_unused2.dataclasses import (
    ReferenceType,
    TemplateReference,
    Template,
    AnalysisResult,
)

init(autoreset=True)

reference_type_colors = {
    ReferenceType.include: Fore.CYAN,
    ReferenceType.extends: Fore.LIGHTBLUE_EX,
    ReferenceType.unknown: Fore.RED,
    ReferenceType.render: Fore.MAGENTA,
}


def print_unreferenced_templates(analysis_result: AnalysisResult, base_dir: str):
    templates = analysis_result.never_referenced_templates
    references = analysis_result.references
    if not templates:
        print(Fore.GREEN + "No unreferenced templates found.")
        return

    # Group templates by AppConfig name
    grouped_templates = {}
    for template in templates:
        if not template.local_app:
            continue
        app_config_name = (
            template.app_config.name if template.app_config else "No AppConfig"
        )
        if app_config_name not in grouped_templates:
            grouped_templates[app_config_name] = []
        grouped_templates[app_config_name].append(template)

    # Print the grouped templates
    print(Fore.YELLOW + "\nTemplates Never Referenced by Python files:")
    for app_config_name, templates in grouped_templates.items():
        print(Fore.CYAN + f"\nAppConfig: {app_config_name}")
        for template in templates:
            print(Fore.MAGENTA + f"{template.relative_path}")
            print_referenced_by(
                template, references, templates_by_id=analysis_result.templates_by_id
            )


def print_referenced_by(
    template: Template,
    references: List[TemplateReference],
    templates_by_id: Dict[str, Template],
):
    reference_chains = get_reference_chain(
        template_id=template.id, references=references, chain=[]
    )

    def rel_path(template_reference: TemplateReference) -> str:
        relative_path = templates_by_id[template_reference.source_id].relative_path
        return f"{Fore.WHITE}{relative_path} ({reference_type_colors.get(template_reference.reference_type, Fore.WHITE)}{template_reference.reference_type.value}{Fore.WHITE} at {Fore.GREEN}{template_reference.line}{Fore.WHITE})"

    if len(reference_chains) > 0:
        print("Referenced by:\n")
        for rc in reference_chains:
            print("\t -> " + " -> ".join([rel_path(chain) for chain in rc]))
        print()


def get_reference_chain(
    template_id: str,
    references: List[TemplateReference],
    chain: List[TemplateReference],
) -> List[List[TemplateReference]]:
    direct_referrals = [
        r for r in references if r.target_id == template_id and r not in chain
    ]

    results = []
    for dr in direct_referrals:
        results.extend(
            get_reference_chain(
                template_id=dr.source_id, references=references, chain=chain + [dr]
            )
        )
    if not results and chain:
        return [chain]
    return results


def print_broken_references(references: List[TemplateReference], base_dir: str):
    if not references:
        print(Fore.GREEN + "No broken references found.")
        return

    print(Fore.RED + "\nBroken References Found:")
    for ref in references:
        source_path = os.path.relpath(ref.source_id, base_dir)
        target_path = os.path.relpath(ref.target_id, base_dir)
        ref_type_color = reference_type_colors.get(ref.reference_type, Fore.WHITE)
        print(
            Fore.BLUE
            + f"{source_path} "
            + ref_type_color
            + f"{ref.reference_type.value} "
            + Fore.BLUE
            + f"{target_path} "
            + Fore.YELLOW
            + "at line "
            + Fore.GREEN
            + f"{ref.line}"
        )



import enum
import os
from dataclasses import dataclass, field
from typing import Optional, Iterable, List, ClassVar, Dict

from django.apps import AppConfig


@dataclass
class StringWithLine:
    value: str
    line: int


@dataclass
class TemplateFilterOptions:
    excluded_apps: Optional[Iterable[str]] = None
    excluded_template_dirs: Optional[Iterable[str]] = None
    excluded_templates: Optional[Iterable[str]] = None


class ReferenceType(enum.Enum):
    include = "include"
    extends = "extends"
    unknown = "unknown"
    render = "render"


@dataclass
class TemplateTokenReference:
    file_path: str
    line_number: int
    reference_type: ReferenceType


@dataclass
class TemplateReference:
    source_id: str
    target_id: str
    reference_type: ReferenceType
    """
    Starting with 1
    """
    line: int
    broken: bool = field(default=False)


@dataclass
class BasePath:
    id: str
    base_dir: str
    relative_path: str
    app_config: Optional["AppConfig"]
    local_app: bool

    @property
    def absolute_path(self):
        return os.path.join(self.base_dir, self.relative_path)

    @property
    def relative_dir_path(self):
        return os.path.dirname(self.relative_path)


@dataclass
class Python(BasePath):
    extensions: ClassVar[List[str]] = [".py"]


@dataclass
class Template(BasePath):
    extensions: ClassVar[List[str]] = [".html", ".txt"]


@dataclass
class AnalysisResult:
    never_referenced_templates: List[Template] = field(default_factory=list)
    broken_references: List[TemplateReference] = field(default_factory=list)
    references: List[TemplateReference] = field(default_factory=list)
    templates: List[Template] = field(default_factory=list)
    python_files: List[Python] = field(default_factory=list)

    @property
    def unused_local_filenames(self) -> List[str]:
        return [f.relative_path for f in self.never_referenced_templates if f.local_app]

    @property
    def unused_filenames(self) -> List[str]:
        return [f.relative_path for f in self.never_referenced_templates]

    @property
    def templates_by_id(self) -> Dict[str, Template]:
        return {template.id: template for template in self.templates}

    def __bool__(self) -> bool:
        """Return True if the analysis found no issues, False otherwise."""
        return not self.never_referenced_templates and not self.broken_references



import re
from typing import List, Optional

from django.template.base import TokenType, Lexer

from django_unused2.dataclasses import TemplateTokenReference, ReferenceType


def extract_template_reference(token: str) -> Optional[str]:
    include_pattern = re.compile(r"[a-z]+\s*['\" ](?P<file_path>[^'\" ]+)['\"]")

    include_match = include_pattern.search(token)
    if include_match:
        return include_match.group("file_path")

    return token.split(" ")[0]

    return None


def extract_template_references(template_text: str) -> List[TemplateTokenReference]:
    """
    Extract all template references from a template string with its type.
    """
    lexer = Lexer(template_text)

    tokens = lexer.tokenize()
    result = []
    for token in tokens:
        if token.token_type == TokenType.BLOCK and (
            token.contents.startswith("include ")
            or token.contents.startswith("extends ")
        ):
            reference_type: ReferenceType = ReferenceType.unknown
            if token.contents.startswith("include "):
                reference_type = ReferenceType.include
            if token.contents.startswith("extends "):
                reference_type = ReferenceType.extends
            relative_path = extract_template_reference(token.contents)
            if not relative_path:
                raise ValueError(f"Unknown token: {token.contents}")
            result.append(
                TemplateTokenReference(
                    relative_path,
                    line_number=token.lineno,
                    reference_type=reference_type,
                )
            )
    return result



import json
from argparse import ArgumentParser
from typing import Any, Dict

from django.conf import settings
from django.core.management.base import BaseCommand

from django_unused2.filter import TemplateFilterOptions, run_analysis
from django_unused2.graph import generate_dot, generate_cytoscape_json
from django_unused2.output import print_unreferenced_templates, print_broken_references


class Command(BaseCommand):
    help = "Lists all unused template files."

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument(
            "unused_type",
            type=str,
            nargs="?",
            default="templates",
            choices=["templates", "template_graph"],
            help="What to find: templates (default), views, media",
        )
        parser.add_argument(
            "-xa",
            "--excluded-apps",
            type=str,
            nargs="*",
            help="List of apps to exclude from the search",
            dest="excluded_apps",
        )
        parser.add_argument(
            "-xd",
            "--excluded-template-dirs",
            type=str,
            nargs="*",
            help="List of template directories to exclude from the search",
            dest="excluded_template_dirs",
        )
        parser.add_argument(
            "-xt",
            "--excluded-templates",
            type=str,
            nargs="*",
            help="List of specific templates to exclude from the search",
            dest="excluded_templates",
        )
        parser.add_argument(
            "-c",
            "--config",
            type=str,
            help="Path to a JSON configuration file",
            dest="config_path",
        )

    def handle(self, *args: Any, **options: dict[str, Any]):
        unused_type = options["unused_type"]
        filter_options = get_filter_options(options)

        if unused_type == "templates":
            result = run_analysis(filter_options)
            print_unreferenced_templates(
                analysis_result=result, base_dir=settings.BASE_DIR
            )
            print_broken_references(result.broken_references, settings.BASE_DIR)
            if not result:
                exit(1)
        elif unused_type == "template_graph":
            result = run_analysis(filter_options)
            with open(settings.BASE_DIR / "unused.dot", "w") as f:
                f.write(generate_dot(result))
            with open(settings.BASE_DIR / "cytoscape.json", "w") as f:
                f.write(generate_cytoscape_json(result))

        else:
            self.stderr.write(
                self.style.ERROR(
                    f"{unused_type} is not a valid parameter. Valid parameters are templates, views, and media."
                )
            )
            exit(1)


def get_filter_options(options: Dict[str, Any]) -> TemplateFilterOptions:
    config_path = options.get("config_path")

    if config_path:
        with open(config_path, "r", encoding="utf-8") as config_file:
            config_data = json.load(config_file)
            return TemplateFilterOptions(
                excluded_apps=config_data.get("excluded_apps", []),
                excluded_template_dirs=config_data.get("excluded_template_dirs", []),
                excluded_templates=config_data.get("excluded_templates", []),
            )

    return TemplateFilterOptions(
        excluded_apps=options.get("excluded_apps", []),
        excluded_template_dirs=options.get("excluded_template_dirs", []),
        excluded_templates=options.get("excluded_templates", []),
    )



import ast
import os
from typing import List, Optional, Union, TypeVar, Type, Set

from django.apps import AppConfig, apps
from django.conf import settings

from django_unused2.dataclasses import (
    StringWithLine,
    Template,
    Python,
    TemplateReference,
    TemplateTokenReference,
    ReferenceType,
)
from django_unused2.template_util import extract_template_references


def find_app_templates() -> List[Template]:
    templates: List[Template] = []

    for config in apps.get_app_configs():
        local_app = str(config.path).find(str(settings.BASE_DIR)) > -1
        dir_path = os.path.join(str(config.path), "templates")
        templates.extend(
            find_templates_in_directory(
                dir_path, app_config=config, local_app=local_app
            )
        )
    return templates


T = TypeVar(
    "T", bound=Union[Template, Python]
)  # Ensuring T is bound to our defined classes


def find_in_directory(
    dir_path: str,
    cls: Type[T],
    local_app: bool,
    app_config: Optional["AppConfig"] = None,
) -> List[T]:
    extensions = cls.extensions  # Get the associated extensions directly from the class
    found_items: List[T] = []
    for root, _, files in os.walk(dir_path):
        for file in files:
            extension = os.path.splitext(file)[1]
            if extension in extensions:
                file_path = os.path.join(root, file)
                relative_path = os.path.relpath(file_path, dir_path).replace("\\", "/")
                found_items.append(
                    cls(
                        id=file_path,
                        base_dir=dir_path,
                        relative_path=relative_path,
                        app_config=app_config,
                        local_app=local_app,
                    )
                )
    return found_items


def find_templates_in_directory(
    dir_path: str, local_app: bool, app_config: Optional[AppConfig] = None
) -> List[Template]:
    return find_in_directory(
        dir_path, cls=Template, app_config=app_config, local_app=local_app
    )


def find_python_in_directory(
    dir_path: str, local_app: bool, app_config: Optional[AppConfig] = None
) -> List[Python]:
    return find_in_directory(
        dir_path, cls=Python, app_config=app_config, local_app=local_app
    )


def find_global_templates() -> List[Template]:
    templates: List[Template] = []

    if settings.TEMPLATES:
        for template_backend in settings.TEMPLATES:
            for directory in template_backend.get("DIRS", []):
                local_app = directory.find(str(settings.BASE_DIR)) > -1
                templates.extend(
                    find_templates_in_directory(directory, local_app=local_app)
                )

    return templates


def find_py_files(exclude_dirs: Optional[List[str]] = None) -> List[Python]:
    exclude_dirs = exclude_dirs or [os.path.join("example", "server", "tests")]

    python_extensions = ["py"]
    result: List[Python] = []

    for config in apps.get_app_configs():
        local_app = str(config.path).find(str(settings.BASE_DIR)) > -1
        dir_path = str(config.path)
        new_files = []
        for root, dirs, files in os.walk(dir_path):
            if any(exclude_dir in root for exclude_dir in exclude_dirs):
                print(f"excluding: {root}")
                continue
            for file in files:
                filename, extension = os.path.splitext(file)
                if extension[1:] in python_extensions:
                    absolute_path = os.path.join(root, file).replace("\\", "/")
                    new_files.append(
                        Python(
                            id=absolute_path,
                            base_dir=dir_path,
                            relative_path=os.path.relpath(absolute_path, dir_path),
                            app_config=config,
                            local_app=local_app,
                        )
                    )
            result.extend(new_files)
    return result


def get_normalized_path(base_dir: str, path: str):
    if path.startswith("."):
        return os.path.normpath(os.path.join(base_dir, path))
    return os.path.normpath(path)


def find_template_to_template_references(
    templates: List[Template],
) -> List[TemplateReference]:
    results = []

    templates_by_relative_path = {
        template.relative_path: template for template in templates
    }

    for template in templates:
        with open(template.absolute_path, "r") as file:
            contents = file.read()

        template_references: List[TemplateTokenReference] = extract_template_references(
            contents
        )

        for template_reference in template_references:
            reference_normalized_relative = get_normalized_path(
                template.relative_dir_path, template_reference.file_path
            )

            if reference_normalized_relative in templates_by_relative_path:
                target_template = templates_by_relative_path[
                    reference_normalized_relative
                ]
                results.append(
                    TemplateReference(
                        source_id=template.id,
                        target_id=target_template.id,
                        reference_type=template_reference.reference_type,
                        broken=False,
                        line=template_reference.line_number,
                    )
                )
            else:
                results.append(
                    TemplateReference(
                        source_id=template.id,
                        target_id=reference_normalized_relative,
                        reference_type=template_reference.reference_type,
                        broken=True,
                        line=template_reference.line_number,
                    )
                )

    return results


class StringLiteralVisitor(ast.NodeVisitor):
    def __init__(self, suffixes: Set[str] = None):
        self.suffixes = suffixes or {".html", ".txt"}
        self.found_strings: List[StringWithLine] = []

    def visit_Str(self, node: ast.Str) -> None:
        for suffix in self.suffixes:
            if node.s.endswith(suffix):
                self.found_strings.append(StringWithLine(node.s, node.lineno))
        self.generic_visit(node)

    def visit_Constant(self, node: ast.Constant) -> None:
        for suffix in self.suffixes:
            if isinstance(node.value, str) and node.s.endswith(suffix):
                self.found_strings.append(StringWithLine(node.value, node.lineno))
        self.generic_visit(node)


def extract_string_literals(source_code: str) -> List[StringWithLine]:
    tree = ast.parse(source_code)
    visitor = StringLiteralVisitor()
    visitor.visit(tree)
    return visitor.found_strings


def find_python_to_template_references(
    templates: List[Template], python_files: List[Python]
) -> List[TemplateReference]:
    used_templates = []

    for python_file in python_files:
        with open(python_file.absolute_path, "r") as file:
            source = file.read()

        string_references = extract_string_literals(source_code=source)
        for template in templates:
            for string_reference in string_references:
                if string_reference.value == template.relative_path:
                    used_templates.append(
                        TemplateReference(
                            line=string_reference.line,
                            source_id=python_file.absolute_path,
                            target_id=template.absolute_path,
                            reference_type=ReferenceType.render,
                        )
                    )

    return used_templates


def find_all_references(
    templates: List[Template], python_files: List[Python]
) -> List[TemplateReference]:
    return find_python_to_template_references(
        templates, python_files
    ) + find_template_to_template_references(templates)
